import { callFastlyApi } from "./fastly-api.js";
const cache = new WeakMap();
export async function getKVStoreIdForNameMap(fastlyApiContext) {
    const cacheEntry = cache.get(fastlyApiContext);
    let kvStoreNameMap = cacheEntry?.kvStoreNameMap;
    if (kvStoreNameMap != null) {
        return kvStoreNameMap;
    }
    const kvStoreInfos = await getKVStoreInfos(fastlyApiContext);
    kvStoreNameMap = {};
    for (const { id, name } of kvStoreInfos) {
        kvStoreNameMap[name] = id;
    }
    cache.set(fastlyApiContext, { ...cacheEntry, kvStoreNameMap });
    return kvStoreNameMap;
}
export async function getKVStoreIdForName(fastlyApiContext, kvStoreName) {
    const kvStoreNameMap = await getKVStoreIdForNameMap(fastlyApiContext);
    return kvStoreNameMap[kvStoreName] ?? null;
}
function createArrayGetter() {
    return function (fn) {
        return async function (fastlyApiContext, ...args) {
            const results = [];
            let cursor = null;
            while (true) {
                const queryParams = new URLSearchParams();
                if (cursor != null) {
                    queryParams.set('cursor', cursor);
                }
                const endpoint = fn(...args);
                const response = await callFastlyApi(fastlyApiContext, endpoint, queryParams);
                if (response.status !== 200) {
                    throw new Error('Unexpected data format');
                }
                const infos = await response.json();
                for (const item of infos.data) {
                    results.push(item);
                }
                if (infos.meta.next_cursor === undefined) {
                    break;
                }
                cursor = infos.meta.next_cursor;
            }
            return results;
        };
    };
}
const _getKVStoreInfos = createArrayGetter()(() => `/resources/stores/kv`);
export async function getKVStoreInfos(fastlyApiContext) {
    const cacheEntry = cache.get(fastlyApiContext);
    let kvStoreInfos = cacheEntry?.kvStoreInfos;
    if (kvStoreInfos != null) {
        return kvStoreInfos;
    }
    kvStoreInfos = await _getKVStoreInfos(fastlyApiContext);
    cache.set(fastlyApiContext, { ...cacheEntry, kvStoreInfos });
    return kvStoreInfos;
}
export const _getKVStoreKeys = createArrayGetter()((kvStoreId) => `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys`);
export async function getKVStoreKeys(fastlyApiContext, kvStoreName) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        return null;
    }
    return await _getKVStoreKeys(fastlyApiContext, kvStoreId);
}
export async function kvStoreEntryExists(fastlyApiContext, kvStoreName, key) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        return false;
    }
    const endpoint = `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys/${encodeURIComponent(key)}`;
    const response = await callFastlyApi(fastlyApiContext, endpoint, null, { method: 'HEAD' });
    return response.status === 200;
}
const encoder = new TextEncoder();
export async function kvStoreSubmitFile(fastlyApiContext, kvStoreName, key, data) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        throw new Error('KV Store not found');
    }
    const endpoint = `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys/${encodeURIComponent(key)}`;
    const body = typeof data === 'string' ? encoder.encode(data) : data;
    const response = await callFastlyApi(fastlyApiContext, endpoint, null, {
        method: 'PUT',
        headers: {
            'content-type': 'application/octet-stream',
        },
        body,
    });
    if (response.status !== 200) {
        throw new Error(`Submitting item ${key} gave error: ${response.status}`);
    }
}
export async function kvStoreDeleteFile(fastlyApiContext, kvStoreName, key) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        throw new Error('KV Store not found');
    }
    const endpoint = `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys/${encodeURIComponent(key)}`;
    const response = await callFastlyApi(fastlyApiContext, endpoint, null, {
        method: 'DELETE',
    });
    if (response.status < 200 || response.status >= 300) {
        throw new Error(`Deleting item '${key}' gave error: ${response.status}`);
    }
}
